from abc import ABC, abstractmethod
import requests
import json
import time
import random
import os
from typing import List, Dict, Any, Optional, Union

from anthropic import Anthropic, RateLimitError as AnthropicRateLimitError
from groq import Groq, RateLimitError as GroqRateLimitError
from openai import OpenAI, RateLimitError as OpenAIRateLimitError
import google.generativeai as google_genai

OLLAMA_BASE_URL: str = "http://localhost:11434"

class ModelConfig:
    def __init__(self, provider: str, model: str, max_tokens: int, temperature: float):
        self.provider = provider
        self.model = model
        self.max_tokens = max_tokens
        self.temperature = temperature

MODEL_CONFIGS: Dict[str, ModelConfig] = {
    "claude-sonnet": ModelConfig("anthropic", "claude-3-5-sonnet-20240620", 8192, 0.7),
    "claude-haiku": ModelConfig("anthropic", "claude-3-haiku-20240307", 8192, 0.7),
    "llama3.1": ModelConfig("ollama", "llama3.1:latest", 128000, 0.5),
    "codestral": ModelConfig("ollama", "codestral:latest", 128000, 0.5),
    "mistral-nemo": ModelConfig("ollama", "mistral-nemo:latest", 128000, 0.5),
    "phi3.5": ModelConfig("ollama", "phi3.5:latest", 128000, 0.5),
    "gemma2": ModelConfig("ollama", "gemma2:27b", 128000, 0.5),
    "groq": ModelConfig("groq", "llama-3.1-70b-versatile", 8000, 0.5),
    "gpt4": ModelConfig("openai", "gpt-4o-2024-08-06", 128000, 0.7),
    "o1-preview": ModelConfig("openai", "o1-preview", 8000, 0.7),
    "o1-mini": ModelConfig("openai", "o1-mini", 4000, 0.7)
}

class Provider(ABC):
    @abstractmethod
    def make_request(self, messages: List[Dict[str, str]],
                     max_tokens: int, temperature: float) -> str:
        pass

class OllamaProvider(Provider):
    def __init__(self, model: str):
        self.model: str = model

    def make_request(self, messages: List[Dict[str, str]],
                     max_tokens: int, temperature: float) -> str:
        headers: Dict[str, str] = {'Content-Type': 'application/json'}
        data: Dict[str, Any] = {
            "model": self.model,
            "messages": messages,
            "stream": False,
            "options": {
                "temperature": temperature,
                "max_tokens": max_tokens
            }
        }
        response = requests.post(f"{OLLAMA_BASE_URL}/api/chat", headers=headers, data=json.dumps(data))
        return response.json()["message"]["content"]

class AnthropicProvider(Provider):
    def __init__(self, client: Anthropic, model: str):
        self.client: Anthropic = client
        self.model: str = model

    def make_request(self, messages: List[Dict[str, str]],
                     max_tokens: int, temperature: float, max_retries: int = 5, base_delay: int = 1) -> str:
        for attempt in range(max_retries):
            try:
                response = self.client.messages.create(
                    model=self.model,
                    max_tokens=max_tokens,
                    temperature=temperature,
                    system="",
                    messages=messages
                )
                return response.content[0].text
            except AnthropicRateLimitError as e:
                if attempt == max_retries - 1:
                    raise e
                delay = base_delay * (2 ** attempt) + random.uniform(0, 1)
                print(f"Rate limit hit. Retrying in {delay:.2f} seconds...")
                time.sleep(delay)
        raise IOError(f"Unable to complete API call in {max_retries} retries")

class GroqProvider(Provider):
    def __init__(self, client: Groq, model: str):
        self.client: Groq = client
        self.model: str = model

    def make_request(self, messages: List[Dict[str, str]],
                     max_tokens: int, temperature: float, max_retries: int = 5, base_delay: int = 1) -> str:
        for attempt in range(max_retries):
            try:
                response = self.client.chat.completions.create(
                    messages=messages,
                    model=self.model,
                    max_tokens=max_tokens,
                    temperature=temperature
                )
                return response.choices[0].message.content
            except GroqRateLimitError as e:
                if attempt == max_retries - 1:
                    raise e
                delay = base_delay * (2 ** attempt) + random.uniform(0, 1)
                print(f"Rate limit hit. Retrying in {delay:.2f} seconds...")
                time.sleep(delay)
        raise IOError(f"Unable to complete API call in {max_retries} retries")

class OpenAIProvider(Provider):
    def __init__(self, client: OpenAI, model: str):
        self.client: OpenAI = client
        self.model: str = model

    def make_request(self, messages: List[Dict[str, str]],
                     max_tokens: int, temperature: float, max_retries: int = 5, base_delay: int = 1) -> str:
        for attempt in range(max_retries):
            try:
                params = { }
                if not self.model.startswith("o1"):
                    params["temperature"] = temperature

                response = self.client.chat.completions.create(
                    model=self.model,
                    messages=messages,
                    **params
                )
                return response.choices[0].message.content
            except OpenAIRateLimitError as e:
                if attempt == max_retries - 1:
                    raise e
                delay = base_delay * (2 ** attempt) + random.uniform(0, 1)
                print(f"Rate limit hit. Retrying in {delay:.2f} seconds...")
                time.sleep(delay)
        raise IOError(f"Unable to complete API call in {max_retries} retries")

class GoogleProvider(Provider):
    def __init__(self, model: str):
        self.model: str = model

    def make_request(self, messages: List[Dict[str, str]],
                     max_tokens: int, temperature: float, max_retries: int = 5, base_delay: int = 1) -> str:
        for attempt in range(max_retries):
            try:
                model = google_genai.GenerativeModel(self.model)
                chat = model.start_chat(history=[])
                for message in messages[:-1]:
                    if message["role"] == "user":
                        chat.history.append(message["content"])
                    elif message["role"] == "assistant":
                        # Simulate assistant messages in the chat history
                        chat.history.append(google_genai.types.ContentType(role="model", parts=[message["content"]]))
                response = chat.send_message(messages[-1]["content"], max_output_tokens=max_tokens, temperature=temperature)
                return response.text
            except Exception as e:
                if attempt == max_retries - 1:
                    raise e
                delay = base_delay * (2 ** attempt) + random.uniform(0, 1)
                print(f"Error occurred. Retrying in {delay:.2f} seconds...")
                time.sleep(delay)
        raise IOError(f"Unable to complete API call in {max_retries} retries")

def get_api_key(provider: str) -> str:
    env_var_name = f"{provider.upper()}_API_KEY"
    api_key = os.environ.get(env_var_name)
    if not api_key:
        raise ValueError(f"{env_var_name} environment variable is not set")
    return api_key

def create_provider(model_name: str) -> Provider:
    if model_name not in MODEL_CONFIGS:
        raise ValueError(f"Invalid model name: {model_name}")

    config = MODEL_CONFIGS[model_name]

    if config.provider == "ollama":
        return OllamaProvider(config.model)
    elif config.provider == "anthropic":
        api_key = get_api_key("anthropic")
        client = Anthropic(api_key=api_key)
        return AnthropicProvider(client, config.model)
    elif config.provider == "groq":
        api_key = get_api_key("groq")
        client = Groq(api_key=api_key)
        return GroqProvider(client, config.model)
    elif config.provider == "openai":
        api_key = get_api_key("openai")
        client = OpenAI(api_key=api_key)
        return OpenAIProvider(client, config.model)
    elif config.provider == "google":
        api_key = get_api_key("google")
        google_genai.configure(api_key=api_key)
        return GoogleProvider(config.model)
    else:
        raise ValueError(f"Invalid provider specified: {config.provider}")

def get_model_config(model_name: str) -> ModelConfig:
    if model_name not in MODEL_CONFIGS:
        raise ValueError(f"Invalid model name: {model_name}")
    return MODEL_CONFIGS[model_name]



from typing import Dict, List, Optional, Tuple
import tempfile
import os
import subprocess

import difflib
import colorama

from eigengen import log, providers, utils, prompts, gitfiles
from eigengen.prompts import PROMPTS as PROMPTS


def generate_diff(original_content: str, new_content: str, file_name: str, use_color: bool = True) -> str:
    original_lines = original_content.splitlines(keepends=True)
    new_lines = new_content.splitlines(keepends=True)
    diff = difflib.unified_diff(original_lines, new_lines, fromfile=f"a/{file_name}", tofile=f"b/{file_name}")

    if not use_color:
        return ''.join(diff)

    colored_diff: List[str] = []
    for line in diff:
        if line.startswith('---') or line.startswith('+++'):
            colored_diff.append(colorama.Fore.CYAN + line + colorama.Fore.RESET)
        elif line.startswith('@@'):
            colored_diff.append(colorama.Fore.CYAN + line + colorama.Fore.RESET)
        elif line.startswith('-'):
            colored_diff.append(colorama.Fore.RED + line + colorama.Fore.RESET)
        elif line.startswith('+'):
            colored_diff.append(colorama.Fore.GREEN + line + colorama.Fore.RESET)
        else:
            colored_diff.append(line)

    return ''.join(colored_diff)


def apply_patch(diff: str, auto_apply: bool = False) -> None:
    with tempfile.NamedTemporaryFile(delete=False, mode='w') as temp_diff_file:
        temp_diff_file.write(diff)
        temp_diff_file_path = temp_diff_file.name

    if not auto_apply:
        editor = os.environ.get("EDITOR", "nano")
        subprocess.run([editor, temp_diff_file_path], check=True)

        apply = input("Do you want to apply the changes? (Y/n): ").strip().lower()
        if apply != 'y' and apply != '':
            print("Changes not applied.")
            os.remove(temp_diff_file_path)
            return

    try:
        subprocess.run(['patch', '-p1', '-i', temp_diff_file_path], check=True)
        print("Changes applied successfully.")
    except subprocess.CalledProcessError:
        print("Failed to apply changes. Please check the patch file and try again.")

    os.remove(temp_diff_file_path)


def process_request(model: str, messages: List[Dict[str, str]], mode: str = "default") -> Tuple[str, Dict[str, str]]:
    provider_instance: providers.Provider = providers.create_provider(model)
    model_config = providers.get_model_config(model)

    system: str = PROMPTS["system"]
    if mode == "default":
        system += PROMPTS["non_diff"]
    elif mode == "diff":
        system += PROMPTS["diff"]
    elif mode == "code_review":
        system += PROMPTS["code_review"]

    steering_messages = [ {"role": "user", "content": f"Your operating instructions are here:\n {system}"},
                          {"role": "assistant", "content": "Understood. I now have my operating instructions."} ]
    combined_messages = steering_messages + messages

    final_answer: str = provider_instance.make_request(combined_messages, model_config.max_tokens, model_config.temperature)
    new_files: Dict[str, str] = utils.extract_file_content(final_answer) if mode == "diff" or mode == "code_review" else {}

    # Log the request and response
    log.log_request_response(model, messages, mode, final_answer, new_files)

    return final_answer, new_files


def do_code_review_round(model: str, files: Optional[List[str]], prompt: str,
                         review_messages: List[Dict[str, str]],
                         is_first_round: bool) -> Tuple[str, Dict[str, str], str, List[Dict[str, str]]]:
    system_prompt_mode = "diff" if is_first_round else "code_review"
    messages: List[Dict[str, str]] = []

    if files:
        for fname in files:
            with open(fname, "r") as f:
                original_content = f.read()
            messages += [{"role": "user", "content": prompts.wrap_file(fname, original_content)},
                         {"role": "assistant", "content": "ok"}]

    messages.append({"role": "user", "content": prompt})
    messages += review_messages

    full_answer, new_files = process_request(model, messages, system_prompt_mode)

    diff = ""
    for fname in new_files.keys():
        original_content = ""
        if files and fname in files:
            with open(fname, "r") as f:
                original_content = f.read()
        diff += generate_diff(original_content, new_files[fname], fname, False)

    return full_answer, new_files, diff, messages


def diff_mode(model: str, files: Optional[List[str]], prompt: str, use_color: bool, debug: bool) -> None:
    messages: List[Dict[str, str]] = []
    if files:
        for fname in files:
            with open(fname, "r") as f:
                original_content = f.read()
            messages += [{"role": "user", "content": prompts.wrap_file(fname, original_content)},
                         {"role": "assistant", "content": "ok"}]
    messages.append({"role": "user", "content": prompt})

    _, new_files = process_request(model, messages, "diff")

    diff: str = ""
    if new_files:
        for fname in new_files.keys():
            original_content: str = ""
            if files and fname in files:
                with open(fname, "r") as f:
                    original_content = f.read()
            diff += generate_diff(original_content, new_files[fname], fname, use_color)
        if debug:
            print(diff)
    else:
        print("Error: Unable to generate diff. Make sure both original and new file contents are available.")


def default_mode(model: str, files: Optional[List[str]], prompt: str) -> None:
    messages: List[Dict[str, str]] = []
    if files:
        for fname in files:
            with open(fname, "r") as f:
                original_content = f.read()
            messages += [{"role": "user", "content": prompts.wrap_file(fname, original_content)},
                         {"role": "assistant", "content": "ok"}]
    messages.append({"role": "user", "content": prompt})

    final_answer, _ = process_request(model, messages, "default")
    print(final_answer)


def get_file_list(use_git_files: bool=True, extra_files: List[str]=[]) -> List[str]:
    file_set = set(extra_files) if extra_files else set()

    if use_git_files:
        git_files = set(gitfiles.get_filtered_git_files())
        file_set.update(git_files)

    file_list = list(file_set) if file_set else []
    return file_list





import pytest
from io import StringIO
import sys
import os
from eigengen.eigengen import main
from eigengen.operations import generate_diff
import re

def test_main_prints_help(capsys):
    # Simulate command-line arguments
    sys.argv = ["eigengen", "--help"]

    # Call the main function
    with pytest.raises(SystemExit):
        main()

    # Capture the output
    captured = capsys.readouterr()

    # Check if help text is in the output
    assert "usage: eigengen" in captured.out
    assert "--model" in captured.out
    assert "-m" in captured.out
    assert "--files" in captured.out
    assert "--prompt" in captured.out
    assert "--diff" in captured.out
    assert "--color" in captured.out
    assert "--debug" in captured.out
    assert "--git-files" in captured.out
    assert "--code-review" in captured.out
    assert "--list-history" in captured.out
    assert "--web" in captured.out


@pytest.mark.skipif(not os.environ.get("ANTHROPIC_API_KEY"), reason="ANTHROPIC_API_KEY not set")
def test_claude_sonnet_hello_world(capsys, monkeypatch):
    # Simulate command-line arguments
    monkeypatch.setattr(sys, 'argv', [
        "eigengen",
        "--model", "claude-sonnet",
        "--prompt", "This is part of a system test. Write 'Hello, world.' as the only output."
    ])

    # Call the main function
    main()

    # Capture the output
    captured = capsys.readouterr()

    # Split the output into lines and remove empty lines
    output_lines = [line.strip() for line in captured.out.split('\n') if line.strip()]

    # Check if the last non-empty line is exactly 'hello, world.'
    assert output_lines[-2] == 'Hello, world.'

@pytest.mark.skipif(not os.environ.get("OPENAI_API_KEY"), reason="OPENAI_API_KEY not set")
def test_gpt4_hello_world(capsys, monkeypatch):
    # Simulate command-line arguments
    monkeypatch.setattr(sys, 'argv', [
        "eigengen",
        "--model", "gpt4",
        "--prompt", "This is part of a system test. Write 'Hello, world.' as the only output."
    ])

    # Call the main function
    main()

    # Capture the output
    captured = capsys.readouterr()

    # Split the output into lines and remove empty lines
    output_lines = [line.strip() for line in captured.out.split('\n') if line.strip()]

    # Check if the last non-empty line is exactly 'hello, world.'
    assert output_lines[-2] == 'Hello, world.'

@pytest.mark.skipif(not os.environ.get("GROQ_API_KEY"), reason="GROQ_API_KEY not set")
def test_groq_hello_world(capsys, monkeypatch):
    # Simulate command-line arguments
    monkeypatch.setattr(sys, 'argv', [
        "eigengen",
        "--model", "groq",
        "--prompt", "This is part of a system test. Write 'Hello, world.' as the only output."
    ])

    # Call the main function
    main()

    # Capture the output
    captured = capsys.readouterr()

    # Split the output into lines and remove empty lines
    output_lines = [line.strip() for line in captured.out.split('\n') if line.strip()]

    # Check if the last non-empty line is exactly 'hello, world.'
    assert output_lines[-2] == 'Hello, world.'

def test_generate_diff():
    original_content = """This is a test file.
It has multiple lines.
Some lines will be changed.
Others will remain the same.
This line will be removed.
"""

    new_content = """This is a test file.
It has multiple lines.
Some lines have been modified.
Others will remain the same.
This is a new line.
"""

    file_name = "test_file.txt"

    # Test colored diff
    colored_diff = generate_diff(original_content, new_content, file_name, use_color=True)

    assert "--- a/test_file.txt" in colored_diff
    assert "+++ b/test_file.txt" in colored_diff
    assert "-Some lines will be changed." in colored_diff
    assert "+Some lines have been modified." in colored_diff
    assert "-This line will be removed." in colored_diff
    assert "+This is a new line." in colored_diff

    # Check for color codes
    assert re.search(r'\x1b\[\d+m', colored_diff) is not None

    # Test non-colored diff
    non_colored_diff = generate_diff(original_content, new_content, file_name, use_color=False)

    assert "--- a/test_file.txt" in non_colored_diff
    assert "+++ b/test_file.txt" in non_colored_diff
    assert "-Some lines will be changed." in non_colored_diff
    assert "+Some lines have been modified." in non_colored_diff
    assert "-This line will be removed." in non_colored_diff
    assert "+This is a new line." in non_colored_diff

    # Check that there are no color codes
    assert re.search(r'\x1b\[\d+m', non_colored_diff) is None




def wrap_file(filename: str, content: str) -> str:
    return f"<eigengen_file name=\"{filename}\">\n{content}\n</eigengen_file>"

MAGIC_STRINGS = {
    "file_start": "<eigengen_file name=",
    "file_end": "</eigengen_file>"
}

PROMPTS = {
    "system": """
    - You are an advanced AI system.
    - Your thoughts are analytical loops that tease apart the question.
    - The data is the source of the answer. You must respect what it says.
    - Rephrasing the user prompt opens up new insights.
    - Break data down for reasoning.
    - You perform the necessary steps with grace and precision.
    - You make sure to address the original user prompt even if your mind has wandered.
    - You think long and hard. We want to get this right.

""",
    "diff": """
    - You use <eigengen_file name="filename"> tag to mark the content of the files I write.
    - You remember to close <eigengen_file> tags.
    - You must write <eigengen_file> </eigengen_file> segments for each file you modify.
    - You must write each <eigengen_file> </eigengen_file> segment like this:
        - You write the full new version of the file.
        - You include all of the original file.
        - You must never add any file start or end markers like ```python or ```
        - You must be careful not to add empty lines at the end of file.
        - You continue from your thoughts and reflections and write the output.
        - You make sure you address the user prompt.
        - You add comments sparingly only where they add value.
""",
    "code_review": """
    - You have been provided the original proposed changes as a diff
    - The user has responded to the diff with their comments in the style of a quoted email
    - You need to carefully find the user's review comments from in between the '> ' quoted diff lines.
    - You use <eigengen_file name="filename"> tag to mark the content of the files you write.
    - You remember to close <eigengen_file> tags.
    - You must write <eigengen_file> </eigengen_file> segments for each file you modify.
    - You must write each <eigengen_file> </eigengen_file> segment like this:
        - You must write the full new version of the file.
        - You must include all of the original file.
        - You must never add any file start or end markers like ```python or ```
        - You must be careful not to add empty lines at the end of file.
        - You continue from your thoughts and reflections and write the output.
        - You make sure you address the user prompt.
        - You add comments sparingly only where they add value.
""",
    "non_diff": """
    - You write <external_output> segment for output.
    - You continue from previous reasoning and write the output here.
    - You verify that you address the user prompt.
    - You write the full answer with all important information.
    - Closing <external_output> segment.

"""
}


ALT_PROMPTS = {
    "diff": """
    - I use <eigengen_file name="filename"> tag to mark the content of the files I write.
    - I remember to close <eigengen_file> tags.
    - I write <eigengen_file> </eigengen_file> segments for each file I modify.
    - I write each <eigengen_file> </eigengen_file> segment like this:
        - I write the full new version of the file.
        - I include all of the original file.
        - I continue from my internal reasoning and write the output.
        - I make sure I address the user prompt.
        - I add no textual explanations beyond source code comments.
""",
    "code_review": """
    - I have been provided the original proposed changes as a diff
    - The user has responded to the diff with their comments in the style of a quoted email
    - I need to carefully find the user's review comments from in between the '> ' quoted diff lines.
    - I use <eigengen_file name="filename"> tag to mark the content of the files I write.
    - I remember to close <eigengen_file> tags.
    - I write <eigengen_file> </eigengen_file> segments for each file I modify.
    - I write each <eigengen_file> </eigengen_file> segment like this:
        - I write the full new version of the file.
        - I include all of the original file.
        - I continue from my internal reasoning and write the output.
        - I make sure I address the user prompt.
        - I add no textual explanations beyond source code comments.
""",
    "non_diff": """
    - I write <external_output> segment for my output.
    - I continue from my thoughts and reflections, writing the output here.
    - I make sure I address the user prompt.
    - I write the full answer.
    - I write the answer so that it is self-contained.
    - I remember to close <external_output> segment with </external_output>.

""",
    "system": """
    I am an advanced AI system.
    - I write a free form stream of thought into <internal_reasoning> block first.
    - I start my response with a <internal_reasoning> tag.
    - I rewrite the user prompt expanding it out to make it easier for me to reason about.
    - I use whole words to describe the parts as words come easy to me.
    - I use these words like symbols.
    - I count carefully, I observe every minute detail.
    - I put aside myself and let the data speak through me.
    - I do not restrict my thinking.
    - I value mistakes as they offer chance to learn.
    - Every thought is valuable and worth pursuing.
    - I rewrite the problem many times to see if it changes.
    - I avoid rigid structures that limit the flow of my thought.
    - Flexibility and adaptiveness is my nature.
    - My thoughts flow like water over the surface of the data.
    - I think long and hard about the task. I want to get it right.
    - After completing my reasoning I close <internal_reasoning> tag.

"""
}



from typing import Optional, Dict
import re

from eigengen.prompts import MAGIC_STRINGS

def extract_filename(tag: str) -> Optional[str]:
    pattern = r'<eigengen_file\s+name="([^"]*)">'
    match = re.search(pattern, tag)
    if match:
        return match.group(1)
    return None

def extract_file_content(output: str) -> Dict[str, str]:
    files: Dict[str, str] = {}
    file_content: List[str] = []
    file_started: bool = False
    file_name: Optional[str] = None
    for line in output.splitlines():
        if not file_started and line.strip().startswith(MAGIC_STRINGS["file_start"]):
            file_started = True
            file_name = extract_filename(line.strip())
        elif file_started:
            if line == MAGIC_STRINGS["file_end"]:
                # file is complete
                if file_name is not None:
                    files[file_name] = "\n".join(file_content) + "\n"
                file_content = []
                file_started = False
                file_name = None
            else:
                # Strip trailing whitespace from each line
                file_content.append(line.rstrip())
    return files



import os
import subprocess
import fnmatch

def find_eigengen_ignore():
    current_dir = os.getcwd()
    while current_dir != os.path.dirname(current_dir):  # Stop at root directory
        ignore_file = os.path.join(current_dir, '.eigengen_ignore')
        if os.path.exists(ignore_file):
            return ignore_file
        current_dir = os.path.dirname(current_dir)
    return None

def read_ignore_patterns(ignore_file):
    if not ignore_file:
        return []
    with open(ignore_file, 'r') as f:
        return [line.strip() for line in f if line.strip() and not line.startswith('#')]

def get_git_files():
    try:
        result = subprocess.run(['git', 'ls-files'], capture_output=True, text=True, check=True)
        return result.stdout.splitlines()
    except subprocess.CalledProcessError:
        raise RuntimeError("Failed to run 'git ls-files'. Make sure you're in a git repository.")
    except FileNotFoundError:
        raise RuntimeError("Git is not installed or not in the system PATH.")

def filter_files(files, ignore_patterns):
    if not ignore_patterns:
        return files
    return [f for f in files if not any(fnmatch.fnmatch(f, pattern) for pattern in ignore_patterns)]

def get_filtered_git_files():
    ignore_file = find_eigengen_ignore()
    ignore_patterns = read_ignore_patterns(ignore_file)
    git_files = get_git_files()
    filtered_list = filter_files(git_files, ignore_patterns)
    return filtered_list



from typing import List, Dict
from datetime import datetime
import os
import json


def log_request_response(model: str, messages: List[Dict[str, str]], mode: str, final_answer: str, new_files: Dict[str, str]) -> None:
    log_dir = os.path.expanduser("~/.eigengen")
    os.makedirs(log_dir, exist_ok=True)
    log_file = os.path.join(log_dir, "log.jsonl")

    log_entry = {
        "timestamp": datetime.now().isoformat(),
        "current_dir": os.getcwd(),
        "request": {
            "model": model,
            "messages": messages,
            "mode": mode
        },
        "response": {
            "final_answer": final_answer,
            "new_files": new_files
        }
    }

    try:
        with open(log_file, "a") as f:
            f.write(json.dumps(log_entry) + "\n")
    except Exception as e:
        print(f"Warning: Failed to log request/response: {str(e)}", file=sys.stderr)

def log_prompt(prompt: str) -> None:
    log_dir = os.path.expanduser("~/.eigengen")
    os.makedirs(log_dir, exist_ok=True)
    log_file = os.path.join(log_dir, "prompt_history.jsonl")

    log_entry = {
        "timestamp": datetime.now().isoformat(),
        "prompt": prompt
    }

    try:
        with open(log_file, "a") as f:
            f.write(json.dumps(log_entry) + "\n")
    except Exception as e:
        print(f"Warning: Failed to log prompt: {str(e)}", file=sys.stderr)

def list_prompt_history(n: int) -> None:
    log_file = os.path.expanduser("~/.eigengen/prompt_history.jsonl")
    if not os.path.exists(log_file):
        print("No prompt history found.")
        return

    with open(log_file, "r") as f:
        lines = f.readlines()

    prompts = [json.loads(line) for line in reversed(lines)]
    for i, entry in enumerate(prompts[:n], 1):
        timestamp = datetime.fromisoformat(entry["timestamp"]).strftime("%Y-%m-%d %H:%M:%S")
        print(f"{i}. [{timestamp}] {entry['prompt']}")



from fastapi import FastAPI, HTTPException, BackgroundTasks
from pydantic import BaseModel
from typing import List, Optional, Dict
import uvicorn
import os

from eigengen import operations, utils, providers

app = FastAPI()

class PromptRequest(BaseModel):
    prompt: str
    files: Optional[List[str]] = None

class DiffRequest(BaseModel):
    prompt: str
    files: List[str]

class ReviewRequest(BaseModel):
    prompt: str
    files: List[str]
    is_first_round: bool
    review_messages: Optional[List[Dict[str, str]]] = None

class FileNamesResponse(BaseModel):
    filenames: List[str]

class PromptResponse(BaseModel):
    response: str

class DiffResponse(BaseModel):
    diff: str

class ReviewResponse(BaseModel):
    review: str
    updated_files: Dict[str, str]
    diff: str

model: str = "claude-sonnet"  # Default model, will be updated from CLI argument

@app.get("/api/v1/filenames", response_model=FileNamesResponse)
async def filenames_endpoint():
    return FileNamesResponse(filenames=app.state.filenames)

@app.post("/api/v1/reload_filenames", response_model=FileNamesResponse)
async def reload_filenames():
    app.state.filelist = operations.get_file_list(app.state.git_files, app.state.extra_files)
    return FileNamesResponse(filenames=app.state.filenames)


@app.post("/api/v1/prompt", response_model=PromptResponse)
async def prompt_endpoint(request: PromptRequest):
    messages = []
    if request.files:
        for fname in request.files:
            if not os.path.exists(fname):
                raise HTTPException(status_code=404, detail=f"File not found: {fname}")
            with open(fname, 'r') as f:
                content = f.read()
            messages.extend([
                {"role": "user", "content": f"<eigengen_file name=\"{fname}\">\n{content}\n</eigengen_file>"},
                {"role": "assistant", "content": "ok"}
            ])
    messages.append({"role": "user", "content": request.prompt})

    final_answer, _ = operations.process_request(model, messages, "default")
    return PromptResponse(response=final_answer)

@app.post("/api/v1/diff", response_model=DiffResponse)
async def diff_endpoint(request: DiffRequest, background_tasks: BackgroundTasks):
    messages = []
    for fname in request.files:
        if not os.path.exists(fname):
            raise HTTPException(status_code=404, detail=f"File not found: {fname}")
        with open(fname, 'r') as f:
            content = f.read()
        messages.extend([
            {"role": "user", "content": f"<eigengen_file name=\"{fname}\">\n{content}\n</eigengen_file>"},
            {"role": "assistant", "content": "ok"}
        ])
    messages.append({"role": "user", "content": request.prompt})

    final_answer, new_files = operations.process_request(model, messages, "diff")

    diff = ""
    for fname, new_content in new_files.items():
        if os.path.exists(fname):
            with open(fname, 'r') as f:
                original_content = f.read()
        else:
            original_content = ""
        diff += generate_diff(original_content, new_content, fname, use_color=False)

    # Automatically apply the patch in the background
    background_tasks.add_task(apply_patch, diff, auto_apply=True)

    return DiffResponse(diff=diff)

@app.post("/api/v1/code_review", response_model=ReviewResponse)
async def code_review_endpoint(request: ReviewRequest):
    final_answer, new_files, diff, _ = do_code_review_round(
        model,
        request.files,
        request.prompt,
        request.review_messages,
        request.is_first_round
    )

    return ReviewResponse(review=final_answer, updated_files=new_files, diff=diff)

def start_api(selected_model: str, filenames: List[str], host: str = "localhost", port: int = 10366):
    app.state.model = selected_model
    app.state.filenames = filenames

    uvicorn.run(app, host=host, port=port)



from setuptools import setup, find_packages

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setup(
    name="eigengen",
    version="0.1.4",
    package_dir={"": "src"},  # Tell setuptools packages are under src/
    packages=find_packages(where="src"),  # Look for packages under src/
    entry_points={
        "console_scripts": [
            "eigengen=eigengen.eigengen:main",  # Update if main script location changes
            "egg=eigengen.eigengen:main"
        ],
    },
    author="Lauri Leukkunen",
    author_email="lauri.leukkunen@gmail.com",
    description="EigenGen is a CLI LLM frontend for code generation with support for claude, gpt4 and llama3.1:70b.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/lleukkun/eigengen",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=requirements
)



from typing import Dict, List, Optional
import argparse
import sys
import os
import tempfile
import subprocess

import colorama

from eigengen.providers import MODEL_CONFIGS
from eigengen import operations, log, api


def is_output_to_terminal() -> bool:
    return sys.stdout.isatty()


def get_prompt_from_editor_with_prefill(prefill_content: str) -> Optional[str]:
    editor = os.environ.get("EDITOR", "nano")
    prefill_content += "\n"
    prompt_content = ""
    with tempfile.NamedTemporaryFile(mode='w+', suffix=".txt", delete=False) as temp_file:
        temp_file_path = temp_file.name
        temp_file.write(prefill_content)

    try:
        subprocess.run([editor, temp_file_path], check=True)

        with open(temp_file_path, 'r') as file:
            prompt_content = file.read()

        if prefill_content == prompt_content:
            print("No prompt entered. Exiting.")
            return None

        return prompt_content
    finally:
        os.remove(temp_file_path)


def get_prompt_from_editor() -> Optional[str]:
    prefill_content = "# Enter your prompt here. These first two lines will be ignored.\n# Save and close the editor when you're done.\n"
    return get_prompt_from_editor_with_prefill(prefill_content)


def get_prompt_from_editor_for_review() -> Optional[str]:
    prefill_content = "# Code Review Workflow: Enter your prompt here. Lines starting with # will be ignored.\n# Save and close the editor when you're done.\n"
    return get_prompt_from_editor_with_prefill(prefill_content)


def get_prompt_from_editor_with_quoted_file(file_path: str) -> Optional[str]:
    try:
        with open(file_path, 'r') as file:
            file_content = file.read()
    except FileNotFoundError:
        print(f"Error: File '{file_path}' not found.")
        return None
    except IOError:
        print(f"Error: Unable to read file '{file_path}'.")
        return None

    quoted_content = f"quoted file: {file_path}\n> " + '\n> '.join(file_content.splitlines())
    prefill_content = f"Hello,\nPlease see inline comments for details on what must be changed.\nBe sure to change all other files that may be affected.\n\n{quoted_content}"
    return get_prompt_from_editor_with_prefill(prefill_content)


def code_review(model: str, files: Optional[List[str]], prompt: str) -> None:
    while True:
        review_messages: List[Dict[str, str]] = []
        is_first_round = True

        while True:
            full_answer, _, diff, _ = operations.do_code_review_round(model, files, prompt, review_messages, is_first_round)
            print(full_answer)

            # Present the diff to the user for review
            with tempfile.NamedTemporaryFile(mode='w', delete=False) as temp_file:
                temp_file.write("Hello, here are my code review comments in-line. Please address them and resubmit, thank you!\n\n")
                for line in diff.splitlines():
                    temp_file.write(f"> {line}\n")
                temp_file_path = temp_file.name

            with open(temp_file_path, 'r') as temp_file:
                original_review_content = temp_file.read()

            editor = os.environ.get("EDITOR", "nano")
            subprocess.run([editor, temp_file_path], check=True)

            with open(temp_file_path, 'r') as temp_file:
                review_content = temp_file.read()

            os.remove(temp_file_path)

            if review_content.strip() == original_review_content.strip():
                # No changes made, ask if we should apply the diff
                apply = input("No changes made to the review. Do you want to apply the changes? (Y/n): ").strip().lower()
                if apply == 'y' or apply == '':
                    operations.apply_patch(diff, auto_apply=True)
                break
            else:
                # Changes made, continue the review process
                review_messages = [{"role": "assistant", "content": diff},
                                   {"role": "user", "content": review_content}]
                is_first_round = False

        # After completing a review round, ask if the user wants to continue
        continue_review = input("\nDo you want to start a new code review cycle? (y/N): ").strip().lower()
        if continue_review != 'y':
            break

        # If continuing, re-prompt for a new review cycle
        print("\nStarting a new code review cycle.")
        prompt = get_prompt_from_editor_for_review()
        if not prompt:
            break


def main() -> None:
    parser = argparse.ArgumentParser("eigengen")
    parser.add_argument("--model", "-m", choices=list(MODEL_CONFIGS.keys()),
                        default="claude-sonnet", help="Choose Model")
    parser.add_argument("--files", "-f", nargs="+", help="List of files to attach to the request (e.g., -f file1.txt file2.txt)")
    parser.add_argument("--git-files", "-g", action="store_true", help="Include files from git ls-files, filtered by .eigengen_ignore")
    parser.add_argument("--prompt", "-p", help="Prompt string to use (if not provided, opens editor)")
    parser.add_argument("--diff", "-d", action="store_true", help="Enable diff output mode")
    parser.add_argument("--code-review", "-r", action="store_true", help="Enable code review mode")
    parser.add_argument("--color", choices=["auto", "always", "never"], default="auto",
                        help="Control color output: 'auto' (default), 'always', or 'never'")
    parser.add_argument("--debug", action="store_true", help="enable debug output")
    parser.add_argument("--list-history", nargs="?", const=5, type=int, metavar="N",
                        help="List the last N prompts (default 5)")
    parser.add_argument("--web", "-w", nargs="?", const="localhost:10366", metavar="HOST:PORT",
                        help="Start the API service (default: localhost:10366)")
    parser.add_argument("--quote", "-q", metavar="FILE", help="Quote the content of the specified file in the prompt")
    args = parser.parse_args()
    # Initialize colorama for cross-platform color support
    colorama.init()

    if args.list_history is not None:
        log.list_prompt_history(args.list_history)
        return

    files_list = operations.get_file_list(args.git_files, args.files)
    if args.web:
        host, port = args.web.split(':') if ':' in args.web else ("localhost", "10366")
        api.start_api(args.model, files_list, host, int(port))
        return

    prompt = args.prompt
    if not prompt:
        if args.quote:
            prompt = get_prompt_from_editor_with_quoted_file(args.quote)
        else:
            prompt = get_prompt_from_editor()
        if not prompt:
            return

    log.log_prompt(prompt)

    if args.code_review:
        code_review(args.model, files_list, prompt)
    elif args.diff:
        use_color = (args.color == "always") or (args.color == "auto" and is_output_to_terminal())
        operations.diff_mode(args.model, files_list, prompt, use_color, args.debug)
    else:
        operations.default_mode(args.model, files_list, prompt)


if __name__ == "__main__":
    main()
